package artifacts

import (
	"context"
	"net/http"
	"net/url"
	"testing"

	"github.com/stretchr/testify/assert"
	testhttp "github.com/stretchr/testify/http"
	"github.com/stretchr/testify/mock"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	kubefake "k8s.io/client-go/kubernetes/fake"

	"github.com/argoproj/argo/persist/sqldb/mocks"
	wfv1 "github.com/argoproj/argo/pkg/apis/workflow/v1alpha1"
	fakewfv1 "github.com/argoproj/argo/pkg/client/clientset/versioned/fake"
	"github.com/argoproj/argo/server/auth"
	authmocks "github.com/argoproj/argo/server/auth/mocks"
	"github.com/argoproj/argo/util/instanceid"
	"github.com/argoproj/argo/workflow/common"
	hydratorfake "github.com/argoproj/argo/workflow/hydrator/fake"
)

func mustParse(text string) *url.URL {
	u, err := url.Parse(text)
	if err != nil {
		panic(err)
	}
	return u
}

func newServer() *ArtifactServer {
	gatekeeper := &authmocks.Gatekeeper{}
	kube := kubefake.NewSimpleClientset()
	instanceId := "my-instanceid"
	wf := &wfv1.Workflow{
		ObjectMeta: metav1.ObjectMeta{Namespace: "my-ns", Name: "my-wf", Labels: map[string]string{
			common.LabelKeyControllerInstanceID: instanceId,
		}},
		Status: wfv1.WorkflowStatus{
			Nodes: wfv1.Nodes{
				"my-node": wfv1.NodeStatus{
					Outputs: &wfv1.Outputs{
						Artifacts: wfv1.Artifacts{
							{
								Name: "my-artifact",
								ArtifactLocation: wfv1.ArtifactLocation{
									Raw: &wfv1.RawArtifact{
										Data: "my-data",
									},
								},
							},
						},
					},
				},
			},
		}}
	argo := fakewfv1.NewSimpleClientset(wf, &wfv1.Workflow{
		ObjectMeta: metav1.ObjectMeta{Namespace: "my-ns", Name: "your-wf"}})
	ctx := context.WithValue(context.WithValue(context.Background(), auth.KubeKey, kube), auth.WfKey, argo)
	gatekeeper.On("Context", mock.Anything).Return(ctx, nil)
	a := &mocks.WorkflowArchive{}
	a.On("GetWorkflow", "my-uuid").Return(wf, nil)
	return NewArtifactServer(gatekeeper, hydratorfake.Noop, a, instanceid.NewService(instanceId))
}

func TestArtifactServer_GetArtifact(t *testing.T) {
	s := newServer()
	r := &http.Request{}
	r.URL = mustParse("/artifacts/my-ns/my-wf/my-node/my-artifact")
	w := &testhttp.TestResponseWriter{}
	s.GetArtifact(w, r)
	assert.Equal(t, 200, w.StatusCode)
	assert.Equal(t, "filename=\"my-artifact.tgz\"", w.Header().Get("Content-Disposition"))
	assert.Equal(t, "my-data", w.Output)
}

func TestArtifactServer_GetArtifactWithoutInstanceID(t *testing.T) {
	s := newServer()
	r := &http.Request{}
	r.URL = mustParse("/artifacts/my-ns/your-wf/my-node/my-artifact")
	w := &testhttp.TestResponseWriter{}
	s.GetArtifact(w, r)
	assert.NotEqual(t, 200, w.StatusCode)
}

func TestArtifactServer_GetArtifactByUID(t *testing.T) {
	s := newServer()
	r := &http.Request{}
	r.URL = mustParse("/artifacts/my-uuid/my-node/my-artifact")
	w := &testhttp.TestResponseWriter{}
	s.GetArtifactByUID(w, r)
	assert.Equal(t, 500, w.StatusCode)
}
